"""
Class containing data in various formats associated with a seismometer (time series, spectra, response).

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
from copy import copy

from gravilab.formats.spectrum import Spectrum


class Seismometer:
    """Class containing data in various formats associated with a seismometer (time series, spectra, response)."""

    def __init__(
        self, time_series=None, spectrum=None, name="generic seismometer", response=None
    ):
        """
        Create object of Seismometer class. Either SeismicTimeSeries or Spectrum should be supplied on input.

        Parameters
        ----------
        name: str
            name for this seismometer, optional
        time_series: SeismicTimeSeries
            time series associated with this seismometer (for each channel)
        spectrum: dict of Spectrum
            spectrum associated with this seismometer (for each channel)
        response:
            pole-zero response dictionary associated with this Seismometer
        """
        if time_series is None and spectrum is None:
            raise ValueError(
                "Must specify either time series (`time_series`) or a spectrum (`spectrum`) on input"
            )
        self._name = name
        self._time_series = time_series
        if spectrum is None:
            self._spectrum = {}
        else:
            self._spectrum = spectrum
        self.response = response

    def __copy__(self):
        """
        Copy this Seismometer object.

        Returns
        -------
        s: Seismometer
            Copy of this Seismometer
        """
        s = Seismometer(
            name=self.name, time_series=self.time_series, spectrum=self._spectrum
        )
        s.response = self.response
        return s

    @property
    def name(self):
        """
        Get seismometer name.

        Returns
        -------
        str
            Name
        """
        return self._name

    def slice_time_series(self, starttime, endtime):
        """
        Slice time series data associated with this element.

        Parameters
        ----------
        starttime: UTCDateTime
            new starting time
        endtime: UTCDateTime
            new end time
        """
        s = copy(self)
        s._time_series = self._time_series.slice(starttime, endtime)
        return s

    @property
    def time_series(self):
        """
        Get time series data associated with this element.

        Returns
        -------
        time_series: TimeSeries
            Time series
        """
        if self._time_series is None:
            raise ValueError("don't have time series data!")
        return self._time_series

    def psd(self, **kwargs):
        """
        Shorthand to calculate default PSD spectrum.

        Parameters
        ----------
        kwargs: dict
            Arguments accepted by spicypy.TimeSeries.psd

        Returns
        -------
        Spectrum
            Spectrum containing pandas DataFrame and some parameters
        """
        return self.spectrum(spectrum_type="psd", **kwargs)

    def asd(self, **kwargs):
        """
        Shorthand to calculate ASD spectrum.

        Parameters
        ----------
        kwargs: dict
            Arguments accepted by spicypy.TimeSeries.psd

        Returns
        -------
        Spectrum
            Spectrum containing pandas DataFrame and some parameters
        """
        return self.spectrum(spectrum_type="asd", **kwargs)

    def spectrum(self, spectrum_type="asd", **kwargs):
        """
        Get time spectrum associated with this element, of specified spectrum type.

        Parameters
        ----------
        spectrum_type: str
            Specify which spectrum type to compute, e.g. "lpsd"

        Returns
        -------
        Spectrum
            Spectrum containing pandas DataFrame and some parameters
        """

        if self._time_series is None:
            raise ValueError(
                'Do not have associated spectrum of type "'
                + spectrum_type
                + '"'
                + " and no time series data to calculate it. Exiting!"
            )

        time_series = self._time_series.to_spicypy()
        spectra_dict = {}
        for channel in time_series:
            if spectrum_type == "psd":  # use default (Welch) averaging in Spicypy
                spectra_dict[channel] = time_series[channel].psd(**kwargs)
            elif spectrum_type == "asd":
                spectra_dict[channel] = time_series[channel].asd(**kwargs)
            else:
                raise ValueError(
                    "Only `asd` and `psd` are supported as spectrum types. "
                    "Perhaps you wanted to specify different averaging method? `method` keyword."
                )

        self._spectrum[spectrum_type] = Spectrum(
            data=spectra_dict, spectrum_type=spectrum_type, response=self.response
        )
        return self._spectrum[spectrum_type]



"""
Class to read data (base class).

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
import os
from obspy import read

from gravilab.formats.seismic_time_series import SeismicTimeSeries


class DataReader:
    """Class to read seismic data."""

    def __init__(self, data_path, channels=None, ignore_extension=False):
        """
        Create DataReader object.

        Parameters
        ----------
        data_path: str
            A valid path pointing to a data file or a folder with files (contained data files are found recursively).
        channels: dict
            Channels for data, optional (e.g. an axis along which measurement is performed).
            Contains channel name that should be present in miniseed file.
            If not provided, lumping all data files together assuming one channel "all".
            Example: {"vertical":"HHZ"}
        ignore_extension: bool
            Try to open any files present; by default relies on known seismic data file extensions.
        """
        self._channels = channels
        if channels is None:
            self._channels = {"all": "*"}
        self._data_path = data_path
        self._ignore_extension = ignore_extension
        self._one_file = False
        self._check_path()
        self.data_file_list = []

    def read_data(self):
        """
        Read seismic data from the path specified on init.

        Returns
        -------
        SeismicTimeSeries
            time series data in SeismicTimeSeries class.
        """
        self._find_files()
        data = {}
        print("Reading data...")
        for file in self.data_file_list:
            stream = read(file)
            for channel in self._channels:
                channel_mask = self._channels[channel]
                for trace in stream:
                    if channel_mask == trace.stats.channel or channel_mask == "*":
                        if channel not in data:
                            data[channel] = trace
                        else:
                            data[channel] += trace
        if not data:
            raise ValueError("Could not find any data for channels specified")
        for channel in self._channels:
            if channel not in data:
                raise ValueError("Could not find any data for channel" + channel)
        print("Done.")
        return SeismicTimeSeries(data)

    def _is_seismic_data(self, file):
        """
        Check if file extension is known to be seismic data.
        With self._ignore_extension set to True always returns True

        Returns
        -------
        bool
            Whether tested file is seismic data.
        """
        if self._ignore_extension:
            return True
        if ".msd" in file or ".miniseed" in file or ".mseed" in file:
            return True
        return False

    def _find_files(self):
        """Find seismic data in path specified in init. Recursively checks all files."""
        self.data_file_list = []
        if self._one_file and self._is_seismic_data(self._data_path):
            self.data_file_list = [self._data_path]
        else:
            for root, _, files in os.walk(self._data_path):
                for file in files:
                    current_file = os.path.join(root, file)
                    if self._is_seismic_data(current_file):
                        self.data_file_list.append(current_file)

        if len(self.data_file_list) == 0:
            raise ValueError("ERROR: found no seismic data files. Exiting!")

    def _check_path(self):
        """Check input path and determine whether it's just one file or a folder."""
        # first deal with exception
        if not os.path.isdir(self._data_path) and not os.path.isfile(self._data_path):
            # invalid path (not a file and not a folder)
            raise ValueError(
                'ERROR: Could open path "'
                + str(self._data_path)
                + '" Check if this is a valid path. Exiting!'
            )
        if os.path.isdir(self._data_path):
            self._one_file = False
        elif os.path.isfile(self._data_path):
            self._one_file = True
        else:
            # invalid path (not a file and not a folder)
            raise ValueError(
                'ERROR: Could open path "'
                + str(self._data_path)
                + '" Check if this is a valid path. Exiting!'
            )



"""
Class to hold spectrum data and associated parameters and methods.

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""

from scipy import signal
import numpy as np


class Spectrum:
    """
    Class to hold spectrum data and associated parameters and methods.

    Parameters
    ----------
    data: dict
        Spectrum data: dict of spicypy.FrequencySeries
    spectrum_type: str
        Either "psd" or "asd"
    response:
        pole-zero response dictionary associated with this Seismometer
    """

    def __init__(
        self,
        data,
        spectrum_type="asd",
        response=None,
    ):
        self._data = {}
        self._original_data = data
        for channel in data:
            self._data[channel] = data[channel]
        self._spectrum_type = spectrum_type
        self.response = response

    def _convert_to_response(self, response):
        """
        Convert data into different units (typically digital counts to physical units) using user-provided
        pole-zero response model.

        Returns
        -------
        response_data: dict of spicypy.FrequencySeries
           Spectrum data with response applied
        """
        zeros = response["zeros"]
        poles = response["poles"]
        scale_factor = response["scale_factor"]
        omega_exponent = 0
        if "omega_exponent" in response:
            omega_exponent = int(response["omega_exponent"])
        asd = True if self._spectrum_type == "asd" else False

        num_poly_coef, denom_poly_coef = signal.ltisys.zpk2tf(
            zeros, poles, scale_factor
        )

        response_data = {}
        for channel, channel_data in self._data.items():
            f = channel_data.frequencies.value
            omega = f * 2 * np.pi
            _, resp = signal.freqs(num_poly_coef, denom_poly_coef, omega)
            respamp = np.absolute(resp * np.conjugate(resp))
            dc_bin = False
            if omega[0] == 0.0:  # avoid potential divide by zero
                omega = omega[1:]
                dc_bin = True
            omega_factor = np.power(omega, omega_exponent)
            if dc_bin:
                omega_factor = np.insert(omega_factor, 0, 0.0)
            if asd:
                respamp = np.sqrt(respamp)
            response_data[channel] = omega_factor * channel_data / respamp
        return response_data

    def data(self, response=None):
        """
        Return associated data in spicypy.FrequencySeries.

        Returns
        -------
        data: dict of spicypy.FrequencySeries
           Spectrum data: spicypy.FrequencySeries.
        """
        if response is None:
            response = self.response
        if response is not None:
            return self._convert_to_response(response)
        return self._data

    def __getitem__(self, channel):
        """
        Return associated data for specified channel by [] operator.

        Returns
        -------
        data: spicypy.FrequencySeries
           Spectrum data in specified channel
        """
        return self.data()[channel]

    def type(self):
        """
        Return spectrum type (e.g. "lpsd").

        Returns
        -------
        str
           Spectrum type
        """
        return self._spectrum_type



"""
Base class to hold time series data and associated parameters and methods.

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
import numpy as np
import pandas as pd

from spicypy.signal.time_series import TimeSeries


class SeismicTimeSeries:
    """Class to hold seismic time series and associated parameters and methods."""

    def __init__(self, data):
        """
        Create SeismicTimeSeries from input seismic data.

        Parameters
        ----------
        data: dict of obspy.Trace
        """
        self._data = data

    def data(self):
        """
        Return associated data.

        Returns
        -------
        data: dict
           Time series data as dict of obspy.Trace objects
        """
        return self._data

    def __getitem__(self, channel):
        """
        Return associated data for specified channel

        Returns
        -------
        data: obspy.Trace
           Time series data as obspy.Trace object
        """
        return self.data()[channel]

    @property
    def sampling_rate(self):
        """
        Get sampling rate for these TimeSeries.

        Returns
        -------
        sample_rate: float
           sampling rate
        """
        sampling_rates = []
        for channel in self._data:
            sampling_rates.append(self._data[channel].stats.sampling_rate)
        sample_rate = np.median(sampling_rates)
        std = np.std(sampling_rates)

        if std / sample_rate > 1e-1:
            raise ValueError(
                "Length of some sampling rates deviates a lot from the median. Some data may be corrupt!\n"
                f"Median sample rate: {sample_rate}, standard deviation: {std}"
            )
        return sample_rate

    def to_spicypy(self):
        """
        Return associated data as a dict of Spicypy TimeSeries.

        Returns
        -------
        ts_dict: dict
           Dictionary of spicypy.TimeSeries.
        """
        ts_dict = {}
        df = self.to_data_frame() * 1.0  # multiplying by 1.0 to convert to float,
        for channel in df:  # spectrum calculation algorithms "don't like" integers
            ts = TimeSeries(
                df[channel], sample_rate=self.sampling_rate, channel=channel
            )
            ts_dict[channel] = ts
        return ts_dict

    def to_data_frame(self):
        """
        Return associated data in a common (DataFrame) format.

        Returns
        -------
        df: pandas::DataFrame
           Time series data: DataFrame with times as index and columns for different channels
        """
        data = self._data
        df = pd.DataFrame()
        index = {}
        for channel in data:
            index[channel] = data[channel].times()
            df.index = data[channel].times()
            df[channel] = data[channel].data

        # sanity check - we expect ALL channels to have the same indices
        if len(index) > 1:
            for _, channel_index in index.items():
                if not np.array_equal(channel_index, df.index):
                    raise ValueError(
                        "ALL channels should have the same time stamps to convert to a DataFrame"
                    )
        return df

    def slice(self, *args, **kwargs):
        """
        Get time series for specified time slices.

        Parameters
        ----------
        args: list
            Positional arguments passed to obspy.Stream()
        kwargs: dict
            Keyword arguments passed to obspy.Stream()

        Returns
        -------
        SeismicTimeSeries
           Sliced time series
        """
        sliced_series = {}
        for channel in self._data:
            st = self._data[channel]
            sliced_series[channel] = st.slice(*args, **kwargs)
        return SeismicTimeSeries(data=sliced_series)

    def __sub__(self, other):
        """
        Subtraction overload.

        Parameters
        ----------
        other: TimeSeries
            Another time series

        Returns
        -------
        TimeSeries
            Subtracted time series
        """
        return self.subtract(other)

    def subtract(self, other_time_series):
        """
        Subtract another time series.

        Parameters
        ----------
        other_time_series: TimeSeries
            Another time series

        Returns
        -------
        TimeSeries
            Subtracted time series
        """
        if not self._match(other_time_series):
            raise ValueError(
                "Time series for subtraction don't match! (Check lengths, channel names etc)"
            )
        subtracted_data = {}
        other_data = other_time_series.data()
        for channel in self._data:
            st1 = self._data[channel].copy()
            st2 = other_data[channel]
            st1.data = st1.data - st2.data
            subtracted_data[channel] = st1
        return SeismicTimeSeries(data=subtracted_data)

    def _match(self, other_time_series):
        """
        Match with another time series, to find out if operations such as subtraction are possible.

        Parameters
        ----------
        other_time_series: TimeSeries
            Another time series

        Returns
        -------
        bool
            Matching or not
        """
        our_data = self._data
        other_data = other_time_series.data()
        for channel in our_data:
            if channel not in other_data:
                print("No channel " + channel + " in other time series")
                return False
        for channel in other_data:
            if channel not in our_data:
                print("No channel " + channel + " in this time series")
                return False
        for channel in our_data:
            if len(our_data[channel]) != len(other_data[channel]):
                print(
                    "Number of traces in our data ("
                    + str(len(our_data[channel]))
                    + ") does not match number of traces in other data"
                    + str(len(other_data[channel]))
                )
                return False
        return True



"""
User interface function to open seismic data files.

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""


from gravilab.input.data_reader import DataReader
from gravilab.base.seismometer import Seismometer


def open(
    data_path,
    channels=None,
    name="generic seismometer",
    response=None,
    ignore_extension=False,
):
    """
    User interface function to open data files.

    Parameters
    ----------
    data_path: str
        A valid path pointing to a data file or a folder with files (contained data files are found recursively)
    channels: dict
        Channels for data, optional (e.g. an axis along which measurement is performed).
        Contains channel name that should be present in miniseed file.
        If not provided, lumping all data files together assuming one channel "all".
    name: str
            name for this seismometer, optional
    response: dict
        Frequency response of the instrument in pole-zero model. Used to convert digital counts to physical units.
        Example: {"poles" =  [-0.036614 +0.037059j, -0.036614 -0.037059j],
                  "zeros" = [0, 0, -31.63, -160, -350, -3177],
                  "scale_factor" = 1202.5 * 8.31871e17 * 400000 #V/(m/s) times gain times counts/V}
    ignore_extension: bool
        Whether to try to open all files independently of extension
        (be careful as any non-seismic files in the folder, such as text files, will lead to a crash).

    Returns
    -------
    Seismometer
       Seismometer object containing time series data
    """
    time_series = DataReader(data_path, channels, ignore_extension).read_data()
    return Seismometer(time_series, name=name, response=response)



"""
Dummy class with same functionality as gravilab.Seismometer - for backwards-compatibility in old scripts

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
from gravilab.base.seismometer import Seismometer


class SystemElement(Seismometer):
    """Dummy class with same functionality as gravilab.Seismometer - for backwards-compatibility in old scripts."""

    def __init__(self, *args, **kwargs):
        """
        Pass arguments to init method of Seismometer class
        Parameters
        ----------
        args: list
            any positional arguments
        kwargs: dict
            any keyword arguments
        """
        super().__init__(*args, **kwargs)



"""
Init and import required modules.

Artem Basalaev <artem[dot]basalaev[at]physik.uni-hamburg.de>,
Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""


import logging as log
from importlib import metadata

from gravilab.open import open
from gravilab.base.seismometer import Seismometer

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    __version__ = "unknown"
    log.warning("Version not known, importlib.metadata is not working correctly.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gravilab', 'gravilab.base', 'gravilab.formats', 'gravilab.input']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19,<2.0',
 'obspy>=1.2.2,<2.0.0',
 'pandas>=1.1,<2.0',
 'spicypy>=0.4.1,<0.5.0']

extras_require = \
{':python_full_version <= "3.7.0"': ['importlib-metadata>=1.1.3']}

setup_kwargs = {
    'name': 'gravilab',
    'version': '0.1.2',
    'description': 'Package to analyze seismic data based on obspy and spicypy packages.',
    'long_description': '# GraviLab\n\nPackage to analyze seismic data based on [obspy](obspy.org) and [spicypy](https://gitlab.com/pyda-group/spicypy) packages.\n\n## Installation\n\n```bash\npip install gravilab\n```\n\n## Usage\n\nThe main element of the package is [`Seismometer`](https://gitlab.com/pyda-group/gravilab/-/blob/main/gravilab/base/seismometer.py) class, which represents a seismic instrument.\nIt can be created by loading seismic data with `open` function and specifying channels,\nboth their names in the file and "human-readable" names. The folder is then scanned for any suitable data\nand only those channels are loaded. All functionality related to seismic time series data is\nprovided by the underlying [obspy](obspy.org) package (specifically [`obspy.Trace`](https://docs.obspy.org/packages/autogen/obspy.core.trace.Trace.html)).\n\n[`Seismometer`](https://gitlab.com/pyda-group/gravilab/-/blob/main/gravilab/base/seismometer.py) class also supports calculating spectrum\n(`Seismometer.asd()`, `Seismometer.psd()`) using [spicypy](https://gitlab.com/pyda-group/spicypy) package.\nSpectrum can be adjusted by the response of specific instrument specified by pole-zero model,\nand units can be converted between displacement, velocity\nand acceleration by specifying the `omega` factor. Resulting spectrum is [`spicypy.FrequencySeries`](https://pyda-group.gitlab.io/spicypy/autoapi/spicypy/signal/frequency_series/index.html). Additionally,\ntime series can also be converted to [`spicypy.TimeSeries`](https://pyda-group.gitlab.io/spicypy/autoapi/spicypy/signal/time_series/index.html) to profit from its rich signal processing functionality.\n\nMinimal example:\n```python\nimport gravilab as gl\nSeismometer1 = gl.open("seismic", channels = {"vertical":"HNZ","north":"HNY","east":"HNX"}, name="seismometer1")\nSeismometer1.time_series["north"].plot()\nresponse = {"poles" : [-0.036614 +0.037059j, -0.036614 -0.037059j, -32.55, -142, -364 +404j, -364 -404j, -1260, -4900 +5200j, -4900 -5200j, -7100 +1700j, -7100 -1700j],\n            "zeros" : [0, 0, -31.63, -160, -350, -3177],\n            "scale_factor" : 1202.5 * 8.31871e17 * 400000, #V/(m/s) times gain times counts/V\n            "omega_exponent": -1, #convert from m/s to m\n           }\nSeismometer1.response = response\nSeismometer1.asd()["north"].plot()\n```\n\nMore detailed usage examples are provided in [gravilab/examples](https://gitlab.com/pyda-group/gravilab/-/tree/main/examples) folder.\n',
    'author': 'Artem Basalaev',
    'author_email': 'artemDOTbasalaev@physikDOTuni-hamburg.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/pyda-group/gravilab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
